#include "ffdb.h"
#include <string.h>
#include <stdlib.h>
#include <new>
#include <assert.h>

namespace ffdb {

//Each row has all its columns stored directly in 'buffer' and not as separate
//values. This makes the memory consumption go down and the whole thing a lot
//faster. The columns are stored properly aligned and padded. NULLs are not
//stored at all in 'buffer'.
//The buffer is allocated with traditional malloc/realloc/free because of the
//realloc feature. Performance-wise it is worth the extra trouble of manually
//throwing std::bad_alloc in case realloc/malloc returns NULL.
//
//The col_ptr containter contains ptr+length items pointing into the buffer


//When the buffer changes location, adjust the pointers for the fields
void Row::adjustptrs(void *oldptr_, void *newptr_) {
	char *oldptr=(char*)oldptr_;
	char *newptr=(char*)newptr_;
	for(container_type::iterator iter=col_ptr.begin();
	    iter!=col_ptr.end();
	    ++iter)
	{
		char *o = (char*)iter->ptr;
		if(o) {
			ptrdiff_t offset = o-oldptr;
			iter->ptr = newptr + offset;
		}
	}
}


Row::Row()
  : col_ptr(),
    buffer_bytes(0),
    buffer_size(0),
    buffer(0)
{
}


Row::Row(const Row &row_)
  : col_ptr(),
    buffer_bytes(0),
    buffer_size(0),
    buffer(0)
{
	*this = row_;
}


Row& Row::operator=(const Row &row_) {
	if(&row_==this) return *this;
	
	void *new_buffer;
	if(row_.buffer) {
		new_buffer = malloc(row_.buffer_bytes);
		if(!new_buffer) throw std::bad_alloc();
	} else
		new_buffer = 0;
	try {
		col_ptr = row_.col_ptr;
	} catch(std::bad_alloc&) {
		free(new_buffer);
		throw;
	}
	free(buffer);
	buffer = new_buffer;
	
	if(row_.buffer) {
		assert(buffer);
		memcpy(buffer,row_.buffer,row_.buffer_bytes);
		buffer_bytes = row_.buffer_bytes;
		buffer_size = row_.buffer_bytes;
		adjustptrs(row_.buffer,buffer);
	} else {
		buffer_bytes=0;
		buffer_size=0;
		buffer=0;
	}
	return *this;
}


void Row::reserve_raw(size_t bytes) {
	if(bytes>buffer_size) {
		if(bytes<8) {
			bytes = (bytes+3)&~3; //round up to a multiple of 4
		} else if(bytes<16) {
			bytes = (bytes+7)&~7; //round up to a multiple of 8
		} else if(bytes<256) {
			bytes = (bytes+15)&~15; //round up to a multiple of 16
		} else
			bytes = (bytes+63)&~63; //round up to a multiple of 64
		
		char *newbuf = (char*)realloc(buffer,bytes);
		if(!newbuf) throw std::bad_alloc();
		if(buffer!=newbuf)
			adjustptrs(buffer,newbuf);
		buffer = newbuf;
		buffer_size = bytes;
	}
}



//Change the number of columns
void Row::resize_columns(size_t columns) {
	col_ptr.resize(columns,VL(0,(size_t)0));
}


//Allocate room for a field with the specified size and the specified 
//alignment requirement.
void *Row::alloc(size_t bytes, size_t alignment) {
	size_t start_offset = buffer_bytes;
	while((start_offset)%alignment)
		start_offset++;
	
	size_t endoffset = start_offset + bytes;
	reserve_raw(endoffset);
	
	buffer_bytes = endoffset;
	void *ptr = (char*)buffer + start_offset;
	col_ptr.push_back(VL(ptr,bytes));
	return ptr;
}


//Allocate room for a field with the specified size and the specified 
//alignment requirement. Then append the value to the row.
void *Row::alloc(size_t bytes, size_t alignment, const void *value) {
	if(value) {
		void *ptr = alloc(bytes,alignment);
		memcpy(ptr,value,bytes);
		return ptr;
	} else {
		col_ptr.push_back(VL(0,(size_t)0));
		return 0;
	}
}

//Append a NULL value to the row.
void Row::alloc_null() {
	col_ptr.push_back(VL(0,(size_t)0));
}

void Row::clear() throw() {
	free(buffer);
	buffer=0;
	buffer_bytes=0;
	buffer_size=0;
	col_ptr.clear(); //lint !e1550
}


bool Row::operator==(const Row &row2) const {
	if(col_ptr.size()!=row2.col_ptr.size()) return false;
	for(size_type i=0; i<col_ptr.size(); i++) {
		const VL &vl1 = col_ptr[i];
		const VL &vl2 = row2.col_ptr[i];
		if(vl1.size!=vl2.size) return false;
		if(vl1.ptr==vl2.ptr) continue; //handles NULL
		if(memcmp(vl1.ptr,vl2.ptr,vl1.size)!=0) return false;
	}
	return true;
}


bool Row::operator!=(const Row &row2) const {
	if(col_ptr.size()!=row2.col_ptr.size()) return true;
	for(size_type i=0; i<col_ptr.size(); i++) {
		const VL &vl1 = col_ptr[i];
		const VL &vl2 = row2.col_ptr[i];
		if(vl1.size!=vl2.size) continue;
		if(vl1.ptr==vl2.ptr) return false; //handles NULL
		if(memcmp(vl1.ptr,vl2.ptr,vl1.size)==0) return false;
	}
	return false;
}


void Row::swap(Row &row2) {
	std::swap(col_ptr,row2.col_ptr);
	std::swap(buffer_bytes,row2.buffer_bytes);
	std::swap(buffer_size,row2.buffer_size);
	std::swap(buffer,row2.buffer);
}





bool is_type_compatible(const RowSetDescriptor &desc1, const RowSetDescriptor &desc2) {
	if(desc1.size()!=desc2.size())
		return false;
	for(columnidx_t i=0; i<desc1.size(); i++) {
		if(desc1[i].type != desc2[i].type)
			return false;
	}
	return true;
}





RowSet::~RowSet() {
}
	
void RowSet::setDescriptor(const RowSetDescriptor &new_desc) {
	desc = new_desc;
	row.clear();
}

bool RowSet::setColumnName(columnidx_t columnidx, const char *name) {
	if(columnidx<desc.size() && name && *name) {
		desc[columnidx].name = name;
		return true;
	} else
		return false;
}

void RowSet::push_back(const Row &new_row) {
	row.push_back(new_row);
}


void RowSet::add_and_swap(Row &new_row) {
	row.push_back(Row());
	back().swap(new_row);
}

} //namespace
