#define __STDC_FORMAT_MACROS
#include "ffdb_io.h"
#include "ffdb_op_alignment.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>

namespace ffdb {

static bool parseDescriptorLine(char *line, RowSet *rowset) {
	RowSetDescriptor new_desc;
	bool first=true;
	char *ss;
	for(char *s=strtok_r(line," \t\r\n",&ss);
	    s;
	    s=strtok_r(0," \t\r\n",&ss))
	{
		if(first) {
			if(strcmp(s,"FFDB")!=0)
				return false;
			first=false;
		} else {
			char *p=strchr(s,':');
			if(!p) return false;
			*p='\0';
			p++;
			switch(*p) {
				case 'i':
					new_desc.push_back(ColumnDescriptor(ColumnType_integer,s));
					break;
				case 'f':
					new_desc.push_back(ColumnDescriptor(ColumnType_real,s));
					break;
				case 's':
					new_desc.push_back(ColumnDescriptor(ColumnType_string,s));
					break;
				case 'r':
					new_desc.push_back(ColumnDescriptor(ColumnType_raw,s));
					break;
				case 'b':
					new_desc.push_back(ColumnDescriptor(ColumnType_boolean,s));
					break;
				default:
					return false;
			}
		}
	}
	rowset->setDescriptor(new_desc);
	return true;
}


static inline void decodehex(void * dst_, size_t /*dstsize*/, const char * src, size_t srcsize) {
	//to be optimized
	uint8_t * dst = (uint8_t*)dst_;
	while(srcsize) {
		char c1 = *src++;
		char c2 = *src++;
		srcsize-=2;
		uint8_t xv1;
		if(c1>='0' && c1<='9') xv1=(uint8_t)(c1-'0');
		else if(c1>='a' && c1<='f') xv1=(uint8_t)(c1-'a'+10);
		else xv1=(uint8_t)(c1-'A'+10);
		uint8_t xv2;
		if(c2>='0' && c2<='9') xv2=(uint8_t)(c2-'0');
		else if(c2>='a' && c2<='f') xv2=(uint8_t)(c2-'a'+10);
		else xv2=(uint8_t)(c2-'A'+10);
		uint8_t value = (uint8_t)((xv1<<4)|xv2);
		*dst++ = value;
	}
}


static inline bool ascii_isdigit(char c) { return c>='0' && c<='9'; }
static inline bool ascii_isspace(char c) { return c==' ' || c=='\t' || c=='\r' || c=='\n'; }
static inline bool ascii_isxdigit(char c) { return (c>='0' && c<='9') || (c>='A' && c<='F') || (c>='a' && c<='f'); }

//strtoll() would work too, but we do not need support for non-10 bases,
//thousands separator, etc. This is much faster. Really.
static inline int64_t strtoint64(const char *str, char **endptr) {
	bool is_negative=false;
	if(*str=='-') {
		str++;
		is_negative=true;
	}
	
	int64_t value=0;
	char c;
	while((c=*str)!='\0') {
		if(!ascii_isdigit(c))
			break;
		str++;
		value = value*10 + (c-'0');
	}
	*endptr = (char*)str;
	if(!is_negative)
		return value;
	else
		return -value;
}


static bool parseDataLine(char *line, size_t &size_hint, RowSet *rowset) {
	const RowSetDescriptor &desc = rowset->descriptor();
	Row new_row;
	new_row.reserve_columns(desc.size());
	if(size_hint<4096) new_row.reserve_raw(size_hint);
	
	RowSetDescriptor::const_iterator desc_iter = desc.begin();
	while(*line) {
		while(ascii_isspace(*line))
			line++;
		if(!*line)
			break;

		if(line[0]=='?' && (line[1]=='\0' || ascii_isspace(line[1]))) {
			//NULL
			line++;
			new_row.alloc_null();
			++desc_iter;
			continue;
		}
		char *n=0;
		
		switch(desc_iter->type) {
			case ColumnType_integer: {
				int64_t value;
				char *endptr=0;
				value = strtoint64(line,&endptr);
				if(endptr) {
					n=endptr;
				} else {
					n=line;
					while(*n) n++;
				}
				
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_integer]);
				*(int64_t*)ptr = value;
				
				break;
			}
			case ColumnType_real:{
				char *endptr=0;
				double value = strtod(line,&endptr);
				if(endptr)
					n=endptr;
				else {
					n=line;
					while(*n) n++;
				}
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_real]);
				*(double*)ptr = value;
				
				break;
			}
			case ColumnType_string:{
				if(*line=='"') {
					std::string value;
					n = line+1;
					for(;;) {
						if(n[0]=='\0') break;
						if(n[0]=='"') break;
						if(n[0]=='\\') {
							n++;
						}
						value += n[0];
						n++;
					}
					if(n[0]=='"') n++;
					void *ptr = new_row.alloc(value.size()+1,ffdb_type_alignment[ColumnType_string],value.data());
					((char*)ptr)[value.size()] = '\0';
				} else {
					n=line;
					while(*n && !ascii_isspace(*n))
						n++;
					size_t sl = (size_t)(n-line);
					
					void *ptr = new_row.alloc(sl+1,ffdb_type_alignment[ColumnType_string],line);
					((char*)ptr)[sl] = '\0';
				}
				break;
			}
			case ColumnType_raw:{
				if(line[0]=='0' && line[1]=='x') line+=2;
				n=line;
				while(ascii_isxdigit(*n))
					n++;
				size_t hexsize = (size_t)(n-line);
				if(hexsize%2) return false;
				size_t rawsize = hexsize/2;
				void *ptr = new_row.alloc(rawsize,ffdb_type_alignment[ColumnType_raw]);
				decodehex(ptr,rawsize,line,hexsize);
				break;
			}
			case ColumnType_boolean: {
				n=line;
				while(*n && !ascii_isspace(*n))
					n++;
				size_t vl=(size_t)(n-line);
				bool value;
				if((vl==4 && strncmp(line,"true",vl)==0) ||
				   (vl==1 && strncmp(line,"T",vl)==0))
					value = true;
				else if((vl==5 && strncmp(line,"false",vl)==0) ||
				        (vl==1 && strncmp(line,"F",vl)==0))
					value = false;
				else
					return false;
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_boolean]);
				*(bool*)ptr = value;
				break;
			}
		}
		assert(n);
		if(*n && !ascii_isspace(*n))
			return false;
		++desc_iter;
		line = n;
	}
	if(new_row.size()!=desc.size())
		new_row.resize_columns(desc.size());
	size_hint = new_row.raw_size();
	
	rowset->add_and_swap(new_row);
	return true;
}


bool ReadFromFile(const char *filename, RowSet *rowset) {
	FILE *fp = fopen(filename,"r");
	if(!fp) return false;
	
	setvbuf(fp,0,_IOFBF,(size_t)65536);
	
	fseek(fp,0L,SEEK_END);
	long file_size = ftell(fp);
	fseek(fp,0L,SEEK_SET);
	
	char line[4096];
	if(fgets(line,(int)sizeof(line),fp)==0) {
		fclose(fp);
		return false;
	}
	
	if(!parseDescriptorLine(line,rowset)) {
		fclose(fp);
		return false;
	}
	
	size_t size_hint=0;
	int row_num=0;
	while(fgets(line,(int)sizeof(line),fp)) {
		if(!parseDataLine(line,size_hint,rowset)) {
			fclose(fp);
			return false;
		}
		row_num++;
		if(row_num==1000) {
			double est_rows = file_size*1000.0/ftell(fp);
			if(est_rows>row_num)
				rowset->reserve((size_t)est_rows);
		}
	}
	fclose(fp);
	return true;
}




static bool writeDescriptor(FILE *fp, const RowSetDescriptor &desc) {
	fprintf(fp,"FFDB");
	for(RowSetDescriptor::const_iterator iter = desc.begin();
	    iter!=desc.end();
	    ++iter)
	{
		fprintf(fp," %s:",iter->name.c_str());
		switch(iter->type) {
			case ColumnType_integer: putc('i',fp); break;
			case ColumnType_real:    putc('f',fp); break;
			case ColumnType_string:  putc('s',fp); break;
			case ColumnType_raw:     putc('r',fp); break;
			case ColumnType_boolean: putc('b',fp); break;
		}
	}
	putc('\n',fp);
	return true;
}


static bool writeRow(FILE *fp, const RowSetDescriptor &desc, const Row &row) {
	Row::const_iterator rowcolumn_iter = row.begin();
	for(columnidx_t i=0; i<desc.size(); i++) {
		if(i!=0) putc(' ',fp);
		const void *value_ptr;
		if(rowcolumn_iter!=row.end())
			value_ptr = rowcolumn_iter->ptr;
		else
			value_ptr = 0;
		if(value_ptr) {
			switch(desc[i].type) {
				case ColumnType_integer: {
					int64_t value = *(const int64_t*)value_ptr;
					fprintf(fp,"%"PRId64, value);
					break;
				}
				case ColumnType_real: {
					double value = *(const double*)value_ptr;
					fprintf(fp,"%f", value);
					break;
				}
				case ColumnType_string: {
					const char *value=(const char *)value_ptr;
					if(*value && strchr(value,' ')==0)
						fputs(value,fp);
					else {
						putc('"',fp);
						char c;
						while((c=*value++)!='\0') {
							if(c=='"' || c=='\\')
								putc('\\',fp);
							putc(c,fp);
						}
						putc('"',fp);
					}
					break;
				}
				case ColumnType_raw: {
					size_t sz=rowcolumn_iter->size;
					if(sz) { //to be optimized
						const uint8_t *value = (const uint8_t*)value_ptr;
						for(; sz; sz--) {
							uint8_t v = *value++;
							fprintf(fp,"%02"PRIx8, v);
						}
					} else
						fputs("0x",fp);
					break;
				}
				case ColumnType_boolean: {
					bool value = *(const bool*)value_ptr;
					fprintf(fp,value?"true":"false");
					break;
				}
			}
		} else {
			putc('?',fp);
		}
		
		if(rowcolumn_iter!=row.end())
			++rowcolumn_iter;
	}
	putc('\n',fp);
	return true;
}


bool WriteToFile(const char *filename, const RowSet &rowset) {
	FILE *fp = fopen(filename,"w");
	if(!fp) return false;
	
	setvbuf(fp,0,_IOFBF,(size_t)65536);
	
	if(!writeDescriptor(fp, rowset.descriptor())) {
		fclose(fp);
		return false;
	}
	for(RowSet::const_iterator iter = rowset.begin();
	    iter!=rowset.end();
	    ++iter)
	{
		if(!writeRow(fp, rowset.descriptor(), *iter)) {
			fclose(fp);
			return false;
		}
	}
	fclose(fp);
	return true;
}


//-----------------------
//CVS (and other formats)


static bool parseDescriptorLine(char *line, RowSet *rowset, header_format_t header_format, char separator) {
	switch(header_format) {
		case header_format_names_only: {
			char sep[4];
			sep[0]=separator;
			sep[1]='\r';
			sep[2]='\n';
			sep[3]='\0';
			columnidx_t colidx=0;
			char *ss;
			for(char *s=strtok_r(line,sep,&ss); s; s=strtok_r(0,sep,&ss)) {
				if(colidx>=rowset->descriptor().size())
					return false; //too many column names in header
				rowset->setColumnName(colidx,s);
				colidx++;
			}
			return true;
		}
	}
	return false;
}


static bool parseDataLine(char *line, size_t size_hint, RowSet *rowset, const char *null_string, char separator) {
	const RowSetDescriptor &desc = rowset->descriptor();
	Row new_row;
	new_row.reserve_columns(desc.size());
	if(size_hint<4096) new_row.reserve_raw(size_hint);
	
	size_t null_string_len = null_string?strlen(null_string):0;
	
	RowSetDescriptor::const_iterator desc_iter = desc.begin();
	while(*line) {
		if(null_string &&
		   strncmp(line,null_string,null_string_len)==0 &&
		   (line[null_string_len]==separator || line[null_string_len]=='\n' || line[null_string_len]=='\0')
		   ) {
			//NULL
			new_row.alloc_null();
			line+=null_string_len;
			if(*line) line++;
			++desc_iter;
			continue;
		}
		char *n=0;
		
		switch(desc_iter->type) {
			case ColumnType_integer: {
				int64_t value;
				char *endptr=0;
				value = strtoint64(line,&endptr);
				if(endptr) {
					n=endptr;
				} else {
					n=line;
					while(*n) n++;
				}
				
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_integer]);
				*(int64_t*)ptr = value;
				
				break;
			}
			case ColumnType_real:{
				char *endptr=0;
				double value = strtod(line,&endptr);
				if(endptr)
					n=endptr;
				else {
					n=line;
					while(*n) n++;
				}
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_real]);
				*(double*)ptr = value;
				
				break;
			}
			case ColumnType_string:{
				if(*line=='"') {
					std::string value;
					n = line+1;
					for(;;) {
						if(n[0]=='\0') break;
						if(n[0]=='"') break;
						if(n[0]=='\\') {
							n++;
						}
						value += n[0];
						n++;
					}
					if(n[0]=='"') n++;
					void *ptr = new_row.alloc(value.size()+1,ffdb_type_alignment[ColumnType_string],value.data());
					((char*)ptr)[value.size()] = '\0';
				} else {
					n=line;
					while(*n && *n!=separator && *n!='\n')
						n++;
					size_t sl = (size_t)(n-line);
					
					void *ptr = new_row.alloc(sl+1,ffdb_type_alignment[ColumnType_string],line);
					((char*)ptr)[sl] = '\0';
				}
				break;
			}
			case ColumnType_raw:{
				if(line[0]=='0' && line[1]=='x') line+=2;
				n=line;
				while(ascii_isxdigit(*n))
					n++;
				size_t hexsize = (size_t)(n-line);
				if(hexsize%2) return false;
				size_t rawsize = hexsize/2;
				void *ptr = new_row.alloc(rawsize,ffdb_type_alignment[ColumnType_raw]);
				decodehex(ptr,rawsize,line,hexsize);
				break;
			}
			case ColumnType_boolean: {
				n=line;
				while(*n && !ascii_isspace(*n))
					n++;
				size_t vl=(size_t)(n-line);
				bool value;
				if((vl==4 && strncmp(line,"true",vl)==0) ||
				   (vl==1 && strncmp(line,"T",vl)==0))
					value = true;
				else if((vl==5 && strncmp(line,"false",vl)==0) ||
				        (vl==1 && strncmp(line,"F",vl)==0))
					value = false;
				else
					return false;
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_boolean]);
				*(bool*)ptr = value;
				break;
			}
		}
		assert(n);
		if(*n && *n!=separator && *n!='\n')
			return false;
		++desc_iter;
		if(*n=='\0' || *n=='\n') break;
		line = n+1;
	}
	if(new_row.size()!=desc.size())
		new_row.resize_columns(desc.size());
	size_hint = new_row.raw_size();
	
	rowset->add_and_swap(new_row);
	return true;
}


bool ReadFromFile(const char *filename, RowSet *rowset, header_format_t header_format, const char *null_string, char separator) {
	FILE *fp = fopen(filename,"r");
	if(!fp) return false;
	
	setvbuf(fp,0,_IOFBF,(size_t)65536);
	
	fseek(fp,0L,SEEK_END);
	long file_size = ftell(fp);
	fseek(fp,0L,SEEK_SET);
	
	char line[4096];
	if(header_format!=header_format_none) {
		if(fgets(line,(int)sizeof(line),fp)==0) {
			fclose(fp);
			return false;
		}
		
		if(!parseDescriptorLine(line,rowset,header_format,separator)) {
			fclose(fp);
			return false;
		}
	}
	
	size_t size_hint=0;
	int row_num=0;
	while(fgets(line,(int)sizeof(line),fp)) {
		if(!parseDataLine(line,size_hint,rowset,null_string,separator)) {
			fclose(fp);
			return false;
		}
		row_num++;
		if(row_num==1000) {
			double est_rows = file_size*1000.0/ftell(fp);
			if(est_rows>row_num)
				rowset->reserve((size_t)est_rows);
		}
	}
	fclose(fp);
	return true;
}


static bool writeDescriptor(FILE *fp, const RowSetDescriptor &desc, header_format_t header_format, char separator) {
	switch(header_format) {
		case header_format_none:
			return true;
		case header_format_names_only:
			for(RowSetDescriptor::const_iterator iter=desc.begin();
			    iter!=desc.end();
			    ++iter)
			{
				if(iter!=desc.begin())
					putc(separator,fp);
				fputs(iter->name.c_str(), fp);
			}
			putc('\n',fp);
			return true;
	}
	return true;
}


static bool writeRow(FILE *fp, const RowSetDescriptor &desc, const Row &row, const char *null_string, char separator, bool quote_strings) {
	Row::const_iterator rowcolumn_iter = row.begin();
	for(columnidx_t i=0; i<desc.size(); i++) {
		if(i!=0) putc(separator,fp);
		const void *value_ptr;
		if(rowcolumn_iter!=row.end())
			value_ptr = rowcolumn_iter->ptr;
		else
			value_ptr = 0;
		if(value_ptr) {
			switch(desc[i].type) {
				case ColumnType_integer: {
					int64_t value = *(const int64_t*)value_ptr;
					fprintf(fp,"%"PRId64, value);
					break;
				}
				case ColumnType_real: {
					double value = *(const double*)value_ptr;
					fprintf(fp,"%f", value);
					break;
				}
				case ColumnType_string: {
					const char *value=(const char *)value_ptr;
					if(*value && strchr(value,' ')==0 && strchr(value,separator) && !quote_strings)
						fputs(value,fp);
					else {
						putc('"',fp);
						char c;
						while((c=*value++)!='\0') {
							if(c=='"' || c=='\\')
								putc('\\',fp);
							putc(c,fp);
						}
						putc('"',fp);
					}
					break;
				}
				case ColumnType_raw: {
					size_t sz=rowcolumn_iter->size;
					if(sz) { //to be optimized
						const uint8_t *value = (const uint8_t*)value_ptr;
						for(; sz; sz--) {
							uint8_t v = *value++;
							fprintf(fp,"%02"PRIx8, v);
						}
					} else
						fputs("0x",fp);
					break;
				}
				case ColumnType_boolean: {
					bool value = *(const bool*)value_ptr;
					fprintf(fp,value?"true":"false");
					break;
				}
			}
		} else {
			if(null_string)
				fputs(null_string,fp);
		}
		
		if(rowcolumn_iter!=row.end())
			++rowcolumn_iter;
	}
	putc('\n',fp);
	return true;

}


bool WriteToFile(const char *filename, const RowSet &rowset, header_format_t header_format, const char *null_string, char separator, bool quote_strings) {
	FILE *fp = fopen(filename,"w");
	if(!fp) return false;
	
	setvbuf(fp,0,_IOFBF,(size_t)65536);
	
	if(!writeDescriptor(fp, rowset.descriptor(), header_format, separator)) {
		fclose(fp);
		return false;
	}
	for(RowSet::const_iterator iter = rowset.begin();
	    iter!=rowset.end();
	    ++iter)
	{
		if(!writeRow(fp, rowset.descriptor(), *iter, null_string, separator, quote_strings)) {
			fclose(fp);
			return false;
		}
	}
	fclose(fp);
	return true;
}

} //namespace
