#include "ffdb_op.h"
#include "ffdb_op_alignment.h"
#include "ffdb_op_cmpdigest.h"
#include <stdlib.h>
#include <stdarg.h>
#include <algorithm>
#include <stdio.h>

namespace ffdb {

bool LockedJoin::prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2) {
	if(cache_valid &&
	  row_set_1.descriptor()==cached_desc1 &&
	  row_set_2.descriptor()==cached_desc2)
		return true;
	cache_valid = false;
	
	cached_desc1 = row_set_1.descriptor();
	cached_desc2 = row_set_2.descriptor();
	cached_desc_result = cached_desc1;
	for(RowSetDescriptor::const_iterator iter=cached_desc2.begin();
	    iter!=cached_desc2.end();
	    ++iter)
		cached_desc_result.push_back(*iter);
	
	if(!BinaryOpHelper::prepare_cache(cached_desc1,cached_desc2))
		return false;
	
	cached_desc2_type_alignment.clear();
	for(RowSetDescriptor::const_iterator iter=cached_desc2.begin();
	    iter!=cached_desc2.end();
	    ++iter)
		cached_desc2_type_alignment.push_back(ffdb_type_alignment[iter->type]);
	
	cache_valid = true;
	
	return true;
}




void LockedJoin::join2rows(const Row &row1, const Row &row2, Row *result) {
	//todo: optimize more
	*result = row1;
	result->reserve_columns(row1.size() + row2.size());
	result->reserve_raw(row1.raw_size() + row2.raw_size());
	size_t i=0;
	for(Row::const_iterator iter=row2.begin(); iter!=row2.end(); ++iter,i++) {
		result->alloc(iter->size,cached_desc2_type_alignment[i],iter->ptr);
	}
}


void LockedJoin::join_rows_with_nulls(RowSet *result)
{
	result->reserve(result->size() + chunk1.rowset_null_rows.size());
	for(RowIndexSet::const_iterator iter = chunk1.rowset_null_rows.begin();
	    iter != chunk1.rowset_null_rows.end();
	    ++iter)
	{
		Row new_row = **iter;
		new_row.resize_columns(cached_desc_result.size());
		result->add_and_swap(new_row);
	}
}



void LockedJoin::execute_sort_merge(RowSet *result)
{
	//  sort digest 1: NlgN
	//  sort digest 2: MlgM
	//  merge-loop: N+M
	//  
	//  =   NlgN + MlgM + N + M
	
	process_rowset_1(true,outer);
	process_rowset_2(true,false);
	
	RowSetDigest::const_iterator i = chunk1.rowset_digest.begin();
	RowSetDigest::const_iterator j = chunk2.rowset_digest.begin();
	Compare_RowDigest cmp;
	while(i!=chunk1.rowset_digest.end() && j!=chunk2.rowset_digest.end()) {
		if(cmp.lt(*i,*j)) {
			if(outer) {
				Row new_row = *i->row_ptr;
				new_row.resize_columns(cached_desc_result.size());
				result->add_and_swap(new_row);
			}
			++i;
		} else if(cmp.gt(*i,*j)) {
			++j;
		} else {
			//join i with [j..{same value}]
			RowSetDigest::const_iterator k=j;
			while(k!=chunk2.rowset_digest.end() && cmp.eq(*i,*k)) {
				Row new_row;
				join2rows(*i->row_ptr, *k->row_ptr, &new_row);
				//result->push_back(new_row);
				result->add_and_swap(new_row);
				++k;
			}
			++i;
		}
	}
}

void LockedJoin::inner_loop_outer(RowSet *result)
{
	//  sort digest 2: MlgM
	//  loop:  N * lgM
	//  =  MlgM + NlgM
	
	process_rowset_2(true,false);
	
	Compare_RowDigest cmp;
	for(RowSetDigest::const_iterator i = chunk1.rowset_digest.begin();
	    i!=chunk1.rowset_digest.end();
	    ++i)
	{
		RowSetDigest::const_iterator j = std::lower_bound(chunk2.rowset_digest.begin(),chunk2.rowset_digest.end(),*i,cmp);
		bool any_joined=false;
		while(j!=chunk2.rowset_digest.end() && cmp.eq(*i,*j)) {
			Row new_row;
			join2rows(*i->row_ptr, *j->row_ptr, &new_row);
			//result->push_back(new_row);
			result->add_and_swap(new_row);
			++j;
			any_joined = true;
		}
		if(!any_joined) {
			Row new_row = *i->row_ptr;
			new_row.resize_columns(cached_desc_result.size());
			result->add_and_swap(new_row);
		}
	}
}


void LockedJoin::inner_loop(RowSet *result)
{
	//  sort digest 2: MlgM
	//  loop:  N * lgM
	//  =  MlgM + NlgM
	
	process_rowset_2(true,false);
	
	Compare_RowDigest cmp;
	for(RowSetDigest::const_iterator i = chunk1.rowset_digest.begin();
	    i!=chunk1.rowset_digest.end();
	    ++i)
	{
		RowSetDigest::const_iterator j = std::lower_bound(chunk2.rowset_digest.begin(),chunk2.rowset_digest.end(),*i,cmp);
		while(j!=chunk2.rowset_digest.end() && cmp.eq(*i,*j)) {
			Row new_row;
			join2rows(*i->row_ptr, *j->row_ptr, &new_row);
			//result->push_back(new_row);
			result->add_and_swap(new_row);
			++j;
		}
	}
}


bool LockedJoin::execute(RowSet *result)
{
	if(!chunk1.rowset || !chunk2.rowset) return false;
	if(!prepare_cache(*chunk1.rowset,*chunk2.rowset)) return false;
	
	process_rowset_1(false,outer);
	process_rowset_2(false,false);
	
	result->setDescriptor(cached_desc_result);
	
	//There was some idea of selecting the lowest cost access method.
	//But since we do not have any disk-I/O to care about inner_loop
	//is always the fastest.
//	if(chunk1.rowset_digest_sorted && chunk2.rowset_digest_sorted) {
//		execute_sort_merge(result);
//	} else if(chunk1.rowset->size() >= chunk2.rowset->size()/2) {
		if(outer)
			inner_loop_outer(result);
		else
			inner_loop(result);
//	} else {
//		execute_sort_merge(result);
//	}
	if(outer) {
		join_rows_with_nulls(result);
	}
	return true;
}


LockedJoin::~LockedJoin() {
}


bool PreparedJoin::execute(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result, bool outer_) {
	lock1(row_set_1);
	lock2(row_set_2);
	setOuter(outer_);
	bool rc = LockedJoin::execute(result);
	unlock1();
	unlock2();
	return rc;
}



bool Join(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
          const char *column_name_1_0, ...)
{
	PreparedJoin prep_join;
	
	va_list ap;
	va_start(ap,column_name_1_0);
	const char *column_name_1 = column_name_1_0;
	while(column_name_1) {
		const char *column_name_2 = va_arg(ap,const char*);
		prep_join.add_relation(column_name_1_0,column_name_2);
		column_name_1 = va_arg(ap,const char*);
	}
	va_end(ap);
	return prep_join.execute(row_set_1,row_set_2,result,false);
}

} //namespace
